import functools
import inspect
import logging
import os
import shutil
import warnings
from collections.abc import Iterable
from dataclasses import asdict
from typing import Any

import torch_geometric.graphgym.register as register
from torch_geometric.data.makedirs import makedirs

try:  # Define global config object
    from yacs.config import CfgNode as CN

    scgnn_cfg = CN()
except ImportError:
    scgnn_cfg = None
    warnings.warn(
        "Could not define global config object. Please install "
        "'yacs' for using the GraphGym experiment manager via "
        "'pip install yacs'."
    )


def set_scgnn_cfg(scgnn_cfg):
    r"""
    This function sets the default config value.
    1) Note that for an experiment, only part of the arguments will be used
    The remaining unused arguments won't affect anything.
    So feel free to register any argument in graphgym.contrib.config
    2) We support *at most* two levels of configs, e.g., scgnn_cfg.dataset.name
    :return: configuration use by the experiment.
    """
    if scgnn_cfg is None:
        return scgnn_cfg

    scgnn_cfg.depth = "all"
    scgnn_cfg.interest_map_norm = True
    scgnn_cfg.score_map_norm = True
    scgnn_cfg.target_baseline = "inference"


def assert_cfg(scgnn_cfg):
    r"""Checks config values, do necessary post processing to the configs
    TODO

    """
    # if scgnn_cfg. not in ["node", "edge", "graph", "link_pred"]:
    # raise ValueError(
    # "Task {} not supported, must be one of node, "
    # "edge, graph, link_pred".format(scgnn_cfg.dataset.task)
    # )
    # scgnn_cfg.run_dir = scgnn_cfg.out_dir


def dump_cfg(scgnn_cfg, path):
    r"""
    TODO
    Dumps the config to the output directory specified in
    :obj:`scgnn_cfg.out_dir`
    Args:
        scgnn_cfg (CfgNode): Configuration node
    """
    makedirs(scgnn_cfg.out_dir)
    scgnn_cfg_file = os.path.join(scgnn_cfg.out_dir, scgnn_cfg.scgnn_cfg_dest)
    with open(scgnn_cfg_file, "w") as f:
        scgnn_cfg.dump(stream=f)


def load_cfg(scgnn_cfg, args):
    r"""
    Load configurations from file system and command line
    Args:
        scgnn_cfg (CfgNode): Configuration node
        args (ArgumentParser): Command argument parser
    """
    scgnn_cfg.merge_from_file(args.scgnn_cfg_file)
    scgnn_cfg.merge_from_list(args.opts)
    assert_scgnn_cfg(scgnn_cfg)


def makedirs_rm_exist(dir):
    if os.path.isdir(dir):
        shutil.rmtree(dir)
    os.makedirs(dir, exist_ok=True)


def get_fname(fname):
    r"""
    Extract filename from file name path
    Args:
        fname (string): Filename for the yaml format configuration file
    """
    fname = fname.split("/")[-1]
    if fname.endswith(".yaml"):
        fname = fname[:-5]
    elif fname.endswith(".yml"):
        fname = fname[:-4]
    return fname


def set_out_dir(out_dir, fname):
    r"""
    Create the directory for full experiment run
    Args:
        out_dir (string): Directory for output, specified in :obj:`scgnn_cfg.out_dir`
        fname (string): Filename for the yaml format configuration file
    """
    fname = get_fname(fname)
    scgnn_cfg.out_dir = os.path.join(out_dir, fname)
    # Make output directory
    if scgnn_cfg.train.auto_resume:
        os.makedirs(scgnn_cfg.out_dir, exist_ok=True)


def set_run_dir(out_dir):
    r"""
    Create the directory for each random seed experiment run
    Args:
        out_dir (string): Directory for output, specified in :obj:`scgnn_cfg.out_dir`
        fname (string): Filename for the yaml format configuration file
    """
    scgnn_cfg.run_dir = os.path.join(out_dir, str(scgnn_cfg.seed))
    # Make output directory
    if scgnn_cfg.train.auto_resume:
        os.makedirs(scgnn_cfg.run_dir, exist_ok=True)


set_scgnn_cfg(scgnn_cfg)


def from_config(func):
    if inspect.isclass(func):
        params = list(inspect.signature(func.__init__).parameters.values())[1:]
    else:
        params = list(inspect.signature(func).parameters.values())

    arg_names = [p.name for p in params]
    has_defaults = [p.default != inspect.Parameter.empty for p in params]

    @functools.wraps(func)
    def wrapper(*args, scgnn_cfg: Any = None, **kwargs):
        if scgnn_cfg is not None:
            scgnn_cfg = (
                dict(scgnn_cfg)
                if isinstance(scgnn_cfg, Iterable)
                else asdict(scgnn_cfg)
            )

            iterator = zip(arg_names[len(args) :], has_defaults[len(args) :])
            for arg_name, has_default in iterator:
                if arg_name in kwargs:
                    continue
                elif arg_name in scgnn_cfg:
                    kwargs[arg_name] = scgnn_cfg[arg_name]
                elif not has_default:
                    raise ValueError(f"'scgnn_cfg.{arg_name}' undefined")
        return func(*args, **kwargs)

    return wrapper
